# -*- coding: utf-8 -*-
# Copyright (c) 2019 NewStore Inc. <engineering@newstore.com>
#
# this file is part of the project "Mercator - Data Mapper for Protobuf" released under the "MIT" open-source license

version = '0.1.8'



import inspect
from .errors import ProtobufCastError
from google.protobuf.descriptor import FieldDescriptor

REGISTRY = {}
BASE_MODEL_CLASS_REGISTRY = {}


class MercatorDomainClass(object):
    """The existence of this class is a trick to avoid redundant imports.

    Any subclasses of this are automatically supported by
    :py:class:`~mercator.meta.FieldMapping` as ``target_type``.

    This was introduced to support
    :py:class:`~mercator.SinglePropertyMapping` but can be used in the
    future in any new types that leverage type casting.
    """


class FieldMapping(object):
    """Base-class for field mapping declaration in :py:class:`~mercator.ProtoMapping`
    that is:

    - :py:class:`~mercator.ProtoKey`
    - :py:class:`~mercator.ProtoList`

    This base-class resides in :py:mod:`mercator.meta`
    so the metaclass can capture the field mapping declarations during
    import-time.

    :param name_at_source: a string with the name of key or property to be extracted in an input object before casting into the target type.
    :param target_type: an optional :py:class:`~mercator.ProtoMapping` subclass or native python type. Check :ref:`target-type` for more details.
    """
    def __init__(self, name_at_source: str, target_type: type = None):
        self.name_at_source = name_at_source
        self.target_type = target_type

        if target_type is not None and not isinstance(target_type, type) and not isinstance(target_type, MercatorDomainClass):
            raise TypeError(f'{self.__class__} takes a type as second argument, but got {type(target_type).__name__} instead')

    def cast(self, value):
        """coerces the given ``value`` into the target type.
        :param value: a python object that is compatible with the given ``target_type``
        """
        if value is None:
            return

        if self.target_type is None:
            return value

        try:
            return self.target_type(value)
        except (ValueError, TypeError) as e:
            msg = str(e)
            raise ProtobufCastError(f'{msg} while casting "{value}" ({type(value).__name__}) to {self.target_type.__name__}')


class ImplicitField(FieldMapping):
    """Like :py:class:`~mercator.ProtoKey` but works is
    declared automagically by the metaclass.
    """


def is_field_property(obj):
    """utility function to check if the give object is a field from a protobuf message.

    For now there is not much metadata when inspecting a protobuf
    class compiled by protoc, so the current strategy is:

    1. Take the ``obj.__class__``
    2. Take the ``__name__`` of the class.
    3. Check that the name contains the string "FieldProperty"

    """
    cls = getattr(obj, '__class__', None)
    name = getattr(cls, '__name__', '')
    return name and 'FieldProperty' in name


def field_properties_from_proto_class(proto_class):
    """inspects all members of the given proto_class and returns a list of
    those who seem to be a message field (determined by :py:meth:`~mercator.meta.is_field_property`)
    """
    members = inspect.getmembers(proto_class)
    return [k for k, v in members if is_field_property(v)]


def validate_proto_attribute(name, attributes):
    """Invoked by :py:class:`~mercator.MetaMapping` during "import time"
    to validate the declaration of :ref:`proto`.
    """
    proto_cls = attributes.get('__proto__')
    if not proto_cls:
        raise SyntaxError(f'class {name} does not define a __proto__')

    return proto_cls

def validate_and_register_base_model_class(cls, name, attributes):
    """Invoked by :py:class:`~mercator.MetaMapping` during "import time"
    to validate the declaration of :ref:`source-input-type`.
    """
    base_model_class = attributes.get('__source_input_type__')
    if not base_model_class:
        return

    if not isinstance(base_model_class, type):
        raise SyntaxError(f'class {name} defined a __source_input_type__ attribute that is not a valid python type: {base_model_class}')
    else:
        BASE_MODEL_CLASS_REGISTRY[base_model_class] = cls


class MetaMapping(type):
    """Metaclass to leverage and enforce correct syntax sugar when
    declaring protomappings.

    Check the source code for comments explaining how everything
    works.
    """
    def __new__(cls, name, bases, attributes):
        cls = type.__new__(cls, name, bases, attributes)
        if name in ('MetaMapping', 'ProtoMapping'):
            return cls

        proto_cls = validate_proto_attribute(name, attributes)
        validate_and_register_base_model_class(cls, name, attributes)

        # extract field names from the __proto__ class, those will
        # become "ImplicitField" instances in the eyes of mercator.
        field_names = field_properties_from_proto_class(proto_cls)

        # create a dictionary with all default implicit fields
        implicit_field_mappings = dict([(k, ImplicitField(k)) for k in field_names])

        # extract all FieldMapping declarations from the ProtoMapping
        # itself, this means all ProtoKey and ProtoList arguments will
        # be considered "explicit fields" in the eyes of mercator.
        explicit_field_mappings = dict([(k, v) for k, v in attributes.items() if isinstance(v, FieldMapping)])

        # store the metadata in the class definition to leverage the
        # whole magic of mapping attributes.
        cls.__field_names__ = field_names
        cls.__implicit_mappings__ = implicit_field_mappings
        cls.__explicit_mappings__ = explicit_field_mappings

        # generate one final dict with the implicit and explicit fields.
        # note the deliberate override of  implicit fields with explicit ones.
        cls.__fields__ = dict(list(implicit_field_mappings.items()) + list(explicit_field_mappings.items()))

        # register all ProtoMapping declarations collected during
        # "import time" in a global dictionary that can be used for
        # instrospection and debugging of existing mappings.
        REGISTRY[name] = cls

        return cls



class TypeCastError(Exception):
    """Raised when trying to cast a value of the wrong type.
    Used primarily by :py:meth:`mercator.ProtoList.cast` from :py:class:`~mercator.ProtoList`
    """


class ProtobufCastError(Exception):
    """raised when failed to create a :py:class:`~google.protobuf.message.Message` instance
    """



# -*- coding: utf-8 -*-
# <Mercator - Python DSL for Protobuf data mapping>
# Copyright (C) <2019>  NewStore Inc <engineering@newstore.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# from google.protobuf.message import Message
from .meta import MetaMapping
from .meta import FieldMapping
from .meta import MercatorDomainClass
# from .meta import BASE_MODEL_CLASS_REGISTRY
from .errors import TypeCastError
from .errors import ProtobufCastError


class SinglePropertyMapping(MercatorDomainClass):
    """creates a new instance of the given protobuf type populated with a
    single property preprocessing the input value with the given callable.

    Example:

    .. code-block:: python
       :emphasize-lines: 8, 13-14

       from mercator import (
           ProtoMapping,
           ProtoKey,
           SinglePropertyMapping,
       )
       from google.protobuf.timestamp_pb2 import Timestamp

       ProtobufTimestamp = SinglePropertyMapping(int, Timestamp, 'seconds')

       class UserAuthTokenMapping(ProtoMapping):
           __proto__ = domain_pb2.User.AuthToken
           value = ProtoKey('data', str)
           created_at = ProtoKey('created_at', ProtobufTimestamp)
           expires_at = ProtoKey('expires_at', ProtobufTimestamp)


        auth_token = UserAuthTokenMapping({'created_at': 12345}).to_protobuf()

        assert isinstance(auth_token.created_at, Timestamp)
        assert auth_token.created_at.seconds == 12345
    """
    def __init__(self, to_python, pb2_type, argname):
        self.to_python = to_python
        self.message_type = pb2_type
        self.argname = argname

    def __call__(self, value):
        params = {}
        input_value = self.to_python(value)
        params[self.argname] = input_value
        return self.message_type(**params)


class ProtoKey(FieldMapping):
    """Represents the intent to translate a object property or dictionary
    key into a protobuf message.

    Use this to map specific values into a protobuf object.

    Example:

    .. code:: python

       class UserMapping(ProtoMapping):
           __proto__ = domain_pb2.User

           username = ProtoKey('login', str)

    :param name_at_source: a string with the name of key or property to be extracted in an input object before casting into the target type.
    :param target_type: an optional :py:class:`~mercator.ProtoMapping` subclass or native python type. Check :ref:`target-type` for more details.
    """
    def cast(self, value):
        """
        :param value: a python object that is compatible with the given ``target_type``
        :returns: ``value`` coerced into the target type. Supports ProtoMappings by automatically calling :py:meth:`~mercator.ProtoMapping.to_protobuf`.
        """
        if value is None:
            return

        result = super().cast(value)
        if not isinstance(result, ProtoMapping):
            return result

        return result.to_protobuf()


class ProtoList(FieldMapping):
    """Represents the intent to translate a several object properties or dictionary
    keys into a list in a protobuf message.


    Example:

    .. code:: python

       class UserMapping(ProtoMapping):
           __proto__ = domain_pb2.User

           tokens = ProtoList('tokens', UserAuthTokenMapping)

    :param name_at_source: a string with the name of key or property to be extracted in an input object before casting into the target type.
    :param target_type: an optional :py:class:`~mercator.ProtoMapping` subclass or native python type. Check :ref:`target-type` for more details.
    """
    def cast(self, value):
        """
        :param value: a python object that is compatible with the given ``target_type``
        :returns: list of items target type coerced into the ``target_type``. Supports ProtoMappings by automatically calling :py:meth:`~mercator.ProtoMapping.to_protobuf`.
        """
        result = super().cast(value)

        if result is None:
            return

        if not isinstance(value, (list, tuple)):
            raise TypeCastError(f'ProtoList.cast() received a non-list value '
                                f'(type {type(value).__name__}): {value}')

        if issubclass(self.target_type, ProtoMapping):
            return [self.target_type(item).to_protobuf() for item in value]

        return [self.target_type(item) for item in value]


def extract_fields_from_dict(data: dict, names: dict):
    """Utility method used by :py:meth:`~mercator.ProtoMapping.to_dict`
    for extracting data plain python dictionaries.

    :param data: a dict with data to be mapped into protobuf objects
    :param names: a :py:class:`dict` with :py:class:`~mercator.meta.FieldMapping` for values.
    :returns: a dict with keyword-arguments to construct new protobuf messages.
    """
    return dict([(name, target.cast(data.get(target.name_at_source))) for name, target in names.items()])


def extract_fields_from_object(data: object, names: dict):
    """Utility method used by :py:meth:`~mercator.ProtoMapping.to_dict`
    for extracting data from instances of :ref:`source-input-type`.

    :param data: a dict with data to be mapped into protobuf objects
    :param names: a :py:class:`dict` with :py:class:`~mercator.meta.FieldMapping` for values.
    :returns: a dict with keyword-arguments to construct new protobuf messages.
    """
    return dict([(name, target.cast(getattr(data, target.name_at_source, None))) for name, target in names.items()])


class ProtoMapping(object, metaclass=MetaMapping):
    """Base class to define attribute mapping from :py:class:`dict` or
    :py:func:`~sqlalchemy.ext.declarative.declarative_base` subclasses'
    instances into pre-filled protobuf messages.

    Example:

    .. code:: python

       from mercator import (
           ProtoMapping,
           ProtoKey,
           ProtoList,
           SinglePropertyMapping,
       )
       from google.protobuf.timestamp_pb2 import Timestamp

       ProtobufTimestamp = SinglePropertyMapping(int, Timestamp, 'seconds')

       class AuthRequestMapping(ProtoMapping):
           __proto__ = domain_pb2.AuthRequest

           username = ProtoKey('username', str)
           password = ProtoKey('password', str)


       class UserAuthTokenMapping(ProtoMapping):
           __proto__ = domain_pb2.User.AuthToken
           value = ProtoKey('data', str)
           created_at = ProtoKey('created_at', ProtobufTimestamp)
           expires_at = ProtoKey('expires_at', ProtobufTimestamp)


       class UserMapping(ProtoMapping):
           __proto__ = domain_pb2.User

           uuid = ProtoKey('id', str)
           email = ProtoKey('email', str)
           username = ProtoKey('login', str)
           tokens = ProtoList('tokens', UserAuthTokenMapping)
           metadata = ProtoKey('extra_info', dict)


       class MediaMapping(ProtoMapping):
           __proto__ = domain_pb2.Media

           author = ProtoKey('author', UserMapping)
           download_url = ProtoKey('link', str)
           blob = ProtoKey('blob', bytes)
           content_type = ProtoKey('content_type', bytes)


       class AuthResponseMapping(ProtoMapping):
           __proto__ = domain_pb2.AuthResponse

           token = ProtoKey('token', UserAuthTokenMapping)

    """
    def __init__(self, data):
        """
        :param data: a :py:class:`dict` or object compatible with the :ref:`source-input-type` declaration at the class level.
        """
        self.data = data

    def to_dict(self):
        """
        :returns: a :py:class:`dict` with keyword-arguments to construct a new instance of protobuf message defined by :ref:`proto`.
        """
        if self.data is None:
            return {}

        fields = self.__fields__
        if isinstance(self.data, dict):
            return extract_fields_from_dict(self.data, fields)

        elif isinstance(self.data, self.__source_input_type__):
            return extract_fields_from_object(self.data, fields)

        else:
            raise TypeError(f'{self.data} must be a dict or {self.__source_input_type__} but is {type(self.data)} instead')

    def to_protobuf(self):
        """
        :returns: a new :ref:`proto` instance with the data extracted with :py:meth:`~mercator.ProtoMapping.to_dict`.
        """
        data = self.to_dict()
        return self.__proto__(**data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"Python DSL to leverage translation of dictionaries and SQLAlchemy into Protobuf objects"


import ast
import os
from setuptools import setup, find_packages


local_file = lambda *f: open(os.path.join(os.path.dirname(__file__), *f)).read()


class VersionFinder(ast.NodeVisitor):
    VARIABLE_NAME = 'version'

    def __init__(self):
        self.version = None

    def visit_Assign(self, node):
        try:
            if node.targets[0].id == self.VARIABLE_NAME:
                self.version = node.value.s
        except:
            pass


def read_version():
    finder = VersionFinder()
    finder.visit(ast.parse(local_file('mercator', 'version.py')))
    return finder.version


def read_requirements():
    return local_file('requirements.txt').splitlines()


def read_test_requirements():
    return local_file('test_requirements.txt').splitlines()


def read_readme():
    """Read README content.
    If the README.rst file does not exist yet
    (this is the case when not releasing)
    only the short description is returned.
    """
    try:
        return local_file('README.rst')
    except IOError:
        return __doc__


setup(
    author='NewStore Inc.',
    author_email='engineering@newstore.com',
    description=read_version(),
    include_package_data=True,
    install_requires=read_requirements(),
    long_description=read_readme(),
    long_description_content_type='text/x-rst',
    name='mercator',
    url='https://mercator.readthedocs.io/en/latest/',
    download_url='https://github.com/NewStore/mercator/releases',
    packages=find_packages(exclude=['*tests*']),
    test_require=read_test_requirements(),
    test_suite='nose.collector',
    version=read_version(),
    package_data={
        'mercator': ' '.join([
            '*.cfg',
            '*.py',
            '*.rst',
            '*.txt',
        ]),
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    zip_safe=False,
)
